#include "stdafx.h"
#include "dics.h"
#include "CommModbusRTU.h"
#include <cstring>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CCommModbusRTU::CCommModbusRTU()
{
	m_sCurData = "";
	m_bRecvCommand = FALSE;
	m_sSendData = "";
}

CCommModbusRTU::~CCommModbusRTU()
{

}

// Recv Frames |Slave Addr|Func Code|Start Addr|Quantity|CRC_LO|CRC_HI|
//             |     1    |    1    |     2    |    2   |   1  |   1  |			// Fixed 8 bytes
//					[0]       [1]      [2][3]    [4][5]    [6]    [7]
// Send Frames |Slave Addr|Func Code|Data bytes|Register Data|CRC_LO|CRC_HI|
//             |     1    |    1    |     1    |      n      |   1  |   1  |	// 5 + n bytes
// Slave addr	: 0x01
// Func code	: 0x03	Read Holding Resgisters
// CRC			: CRC-16 Modbus
void CCommModbusRTU::DataReceived()
{
	if( m_bRecvCommand )
		return;
	else
		m_bRecvCommand = TRUE;

	char buf[256];
	memset(buf, 0x00, 256);
	Sleep( 10 );

	int nLen = XComm::Read( buf, 8 );

	if( nLen == 8 )	// check Data Lenght
	{
		CString sMessage, sTemp;
		for( int i = 0; i < nLen; i++)
		{
			sTemp.Format("%02X ", buf[i]);
			sMessage += sTemp;
		}
		SetLog("ModbusRcv", 0, 0, sMessage);
		
		if( buf[0] == 0x01 )	// check Slave Address
		{
			if( buf[1] == 0x03 || buf[1] == 0x06 )	// check Function code	// 240606 halesu
			{
				//m_sReceivedData = CString( buf );	// parsing ng need to check
				
				// Data Pasing
				int nStartAddress = (UINT)buf[3];
				int nQuantity = (UINT)buf[5];
				unsigned char crc_LO = buf[6]; 
				unsigned char crc_HI = buf[7];

				// Check CRC of Recv Data
				USHORT CRC;
				CRC = CRC16_MODBUS((unsigned char*)buf, 6);
				USHORT CRC_LO = CRC & 0x00FF;
				USHORT CRC_HI = (CRC & 0xFF00) >> 8;

				if( crc_LO != CRC_LO || crc_HI != CRC_HI )
				{
					SetLog("Modbus	", 0, 0, "CRC check NG");
					Reset();
					return;
				}

				// check nStart Address & nQuantity
				int nBytes = 2;
				if( buf[1] == 0x03 )	// read function
				{
					if( nStartAddress < 2 || nStartAddress + nQuantity > MAX_DATA_NO )
					{
						SetLog("Modbus	", 0, 0, "Register Check NG");
						Reset();
						return;
					}
					else 
						nBytes = CalDataBytes( nStartAddress, nQuantity );
				}
				else // write function	// 240606 halesu
				{
					if( nStartAddress != 14 )	// bAlarm
					{
						SetLog("Modbus	", 0, 0, "Register Check NG");
						Reset();
						return;
					}
					else // Set value
						m_pDicsData->m_GenInfo.bScanDeviceAlarmCheck	= buf[5] == 0x30 ? FALSE : TRUE;
				}

				// Update Current Data
				CString sBuf;
				sBuf.Format("%c%c", buf[0], buf[1]);
				m_sCurData = sBuf;
				UpdateData();
				
				if( buf[1] == 0x03 )	
					sBuf.Format("%c", (unsigned char)nBytes);
				else // 240606 halesu
					sBuf.Format("%c%c",  buf[2], buf[3]);

				// Set Send Data 2 + buf + nBytes
				m_sSendData = m_sCurData.Mid(0,2) + sBuf + m_sCurData.Mid(CalDataAddress(nStartAddress), nBytes);

				// Set CRC of Send Data
				CRC = CRC16_MODBUS((unsigned char*)m_sSendData.GetBuffer(), m_sSendData.GetLength());
				CRC_LO = CRC & 0x00FF;
				CRC_HI = (CRC & 0xFF00) >> 8;
				sBuf.Format("%c%c", CRC_LO, CRC_HI);
				m_sSendData += sBuf;

				// Set Log
				CString sMessage, sTemp;
				for( int i = 0; i < m_sSendData.GetLength(); i++)
				{
					sTemp.Format("%02X ", m_sSendData[i]);
					sMessage += sTemp;
				}
				SetLog("ModbusSnd", 0, 0, sMessage);

				// Send Command
				SendCommand();
			}
			else SetLog("Modbus	", 0, 0, "Function Code NG");
		}
		else SetLog("Modbus	", 0, 0, "Slave Address NG");
	}
	else SetLog("Modbus	", 0, 0, "Data Length NG");

	Reset();
}

void CCommModbusRTU::Reset()
{
	m_sCurData = "";
	m_bRecvCommand = FALSE;
	m_sSendData = "";
}
///////////////////////////////
// Combination of m_sCurData
// [0]		nStartAddess
// [1]		nFunctionCode
// [2]~[10]	sGlassId
// [11][12] sPortNo
// [13][14] sSlotNo
// [15][16] sAlarm
//
void CCommModbusRTU::UpdateData()
{
	CString sSheetID;
	CString sPortNo;
	CString sSlotNo;
	CString sSensorOnOff;
	CString sAlarm;
	int nLen = m_pDicsData->m_GlsInfo[IF_UNIT_BUF*2].sGlassId.GetLength();	// 240606 halesu

	if( nLen >= 9 )
		sSheetID = m_pDicsData->m_GlsInfo[IF_UNIT_BUF*2].sGlassId.Mid(0, 9);	// 9 bytes
	else // 240606 halesu
	{
		sSheetID.Format("%-9s", m_pDicsData->m_GlsInfo[IF_UNIT_BUF*2].sGlassId.Mid(0, nLen));	// 9 bytes
	}
	sPortNo.Format("%02d", atoi(m_pDicsData->m_GlsInfo[IF_UNIT_BUF*2].sPortNo));	// 2 bytes
	sSlotNo.Format("%02d", atoi(m_pDicsData->m_GlsInfo[IF_UNIT_BUF*2].sSlotNo));	// 2 bytes
	sSensorOnOff = m_pEqpAssy->m_GlassExistUnit.GetOr( IF_UNIT_BUF*2 + 0 ) ? "01" : "00";	// 2 bytes
	sAlarm = m_pDicsData->m_GenInfo.bScanDeviceAlarmCheck ? "01" : "00";	// 2 bytes	// 240606 halesu

	m_sCurData += sSheetID + sPortNo + sSlotNo + sSensorOnOff + sAlarm;	// 240606 halesu
}
///////////////////////////////////
// ADDR			Data		Bytes
// [2]~[10]		Sheet ID		1
// [11]			CST Port No		2
// [12]			CST Slot No		2
// [13]			3X001(BUF IN)	2
// [14]			Alarm			2
int CCommModbusRTU::CalDataBytes( int nStartAddress, int nQuantity)
{
	int nRv = 0;
	int nDataByte[MAX_DATA_NO] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
								   1, 2, 2, 2, 2,
	};
	
	for (int i = nStartAddress; i < nStartAddress + nQuantity; i++)
		nRv += nDataByte[i];

	return nRv;
}

int CCommModbusRTU::CalDataAddress( int nStartAddress)
{
	int nRv = 0;
	int nDataByte[MAX_DATA_NO] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
								   1, 2, 2, 2, 2,
	};
	
	for (int i = 0; i < nStartAddress; i++)
		nRv += nDataByte[i];

	return nRv;
}

void CCommModbusRTU::SendCommand()
{
	BOOL m_bSendCommand = FALSE;
	
	FlushTx();
	
	m_bSendCommand = Write( (LPCTSTR)m_sSendData );
	if ( !m_bSendCommand )	SetLog("ModbusSnd", 0, 0, "SendCommand NG");

	return;
}

void CCommModbusRTU::SetLog(CString sSeqName, int nPortNo, int nSlotNo, CString sMessage)
{
	// Log
	CString sPortNo;
	if( nPortNo <= 0 ) sPortNo = "";
	else sPortNo.Format( "%d", nPortNo );

	CString sSlotNo;
	if( nSlotNo <= 0 ) sSlotNo = "";
	else sSlotNo.Format( "%d", nSlotNo );

	m_sMessage.Format( "COMM	\t%s\t%s\t%s\t%s", sSeqName, sPortNo, sSlotNo, sMessage );
	Log.TextOut( m_sMessage );

	m_sMessage.Format( "%s : %s", sSeqName, sMessage );
	Msg( WM_EQP_LOG, 0, (LPARAM)(LPCTSTR) m_sMessage );
}

/**
**************************************************************************************************
* @Brief    Single byte data inversion        
* @Param    
*            @DesBuf: destination buffer
*            @SrcBuf: source buffer
* @RetVal    None
* @Note      (MSB)0101_0101 ---> 1010_1010(LSB)
**************************************************************************************************
*/
void CCommModbusRTU::InvertUint8(unsigned char *DesBuf, unsigned char *SrcBuf)
{
    int i;
    unsigned char temp = 0;
    
    for(i = 0; i < 8; i++)
    {
        if(SrcBuf[0] & (1 << i))
        {
            temp |= 1<<(7-i);
        }
    }
    DesBuf[0] = temp;
}

/**
**************************************************************************************************
* @Brief    double byte data inversion        
* @Param    
*            @DesBuf: destination buffer
*            @SrcBuf: source buffer
* @RetVal    None
* @Note      (MSB)0101_0101_1010_1010 ---> 0101_0101_1010_1010(LSB)
**************************************************************************************************
*/
void CCommModbusRTU::InvertUint16(unsigned short *DesBuf, unsigned short *SrcBuf)  
{  
    int i;  
    unsigned short temp = 0;    
    
    for(i = 0; i < 16; i++)  
    {  
        if(SrcBuf[0] & (1 << i))
        {          
            temp |= 1<<(15 - i);  
        }
    }  
    DesBuf[0] = temp;  
}

unsigned short CCommModbusRTU::CRC16_MODBUS(unsigned char *puchMsg, unsigned int usDataLen)  
{  
    unsigned short wCRCin = 0xFFFF;  
    unsigned short wCPoly = 0x8005;  
    unsigned char wChar = 0;
	int i;
    
    while (usDataLen--)     
    {  
        wChar = *(puchMsg++);  
        InvertUint8(&wChar, &wChar);  
        wCRCin ^= (wChar << 8);

        for(i = 0; i < 8; i++)  
        {  
            if(wCRCin & 0x8000) 
            {
                wCRCin = (wCRCin << 1) ^ wCPoly;  
            }
            else  
            {
                wCRCin = wCRCin << 1; 
            }            
        }  
    }  
    InvertUint16(&wCRCin, &wCRCin);  
    return (wCRCin) ;  
} 